var objlen = x => Object.keys(x).length;
var log = x=>{};
if (process.env.model_log)
    log = console.log;

// Validate input options
function validateOptions(o) {
    if (o.action && typeof o.action != 'function')
        throw `"action" must be a function. Got type ${typeof o.action}.`
}

// Return option if required option is missing
function missingOption(options) {
    var required = ['action'];
    for (let item of required) {
        if (!options[item]) {
            return item;
        }
    }
    return false;
}

// Throw exception if name already exists
function checkNameExists(name) {
    if (this.sources[name]) {
       throw `Name ${name} already exists`;
    }
}

// Add a new data source to the model
function add(name, options) {
    this.checkNameExists(name);
    var combined = Object.assign({}, this, options);
    var missing = missingOption(combined);
    if (missing) {
       throw `Required option ${missing} not provided`;
    }
    validateOptions(options);
    // Merge defaults and options
    combined.unions = {};
    combined._name = name;
    this.sources[name] = combined;
    log('[model] Added new data source', name);
}

// Combine multiple sources into one
function union(name, unionOf, action) {
    this.checkNameExists(name);
    var combined = Object.assign({}, this);
    combined.unionOf = unionOf;
    combined.action = action;
    combined.isUnion = true;
    for (let item of unionOf) {
        if (!this.sources[item]) {
            throw `Cannot union ${item}, source does not exist`;
        }
        this.sources[item].unions[name] = combined;
    }
    this.add(name, combined);
    log('[model] Added new union', name, 'from', unionOf);
}

// Join unions into one object
function formUnion(name, results) {
    var union = this.sources[name];
    var data = {};
    for (let key of union.unionOf) {
        if (!results[key]) {
            log('[model] cannot join', name, 'missing', key);
            return;
        }
        data[key] = results[key];
    }
    results[name] = union.action(data);
    log('[model] Joined union', name, 'from', union.unionOf);
}

// Create interval to perform action
function makeGetter(name, build, source) {
    var getter = {source, args:build.args};
    var that = this;
    function run() {
        var setScope = source.action.bind(source);
        source.action(build.args, (err, result) => {
            if (err) {
                console.error('Error from', name, build.args);
                console.error(err);
            } else {
                build.results[name] = result;
                for (var key in source.unions) {
                    that.formUnion(key, build.results);
                }
            }
        });
    }
    run();
    getter.interval = setInterval(x => run, 1000 * 60 * source.minutes);
    build.getters[name] = getter;
    log('[model] Made source connection:', name, 'and', build.args);
}

// Start requests if necessary
function checkBuilds() {
    for (let b of this.builds) {
        for (let key in this.sources) {
            var src = this.sources[key];
            if (src.isUnion || src.isIntersection) {
                log('[model] Skipping', key, 'as it is a union');
                continue;
            }
            if (!b.getters[key]) {
                this.makeGetter(key, b, src);
            }
        }
    }
}

// Specify an argument set
function build(args) {
    var path = Object.keys(args).sort();
    var at = this.buildMap;
    for (let item of path) {
        var value = args[item];
        if (!at[item])
            at[item] = {};
        if (!at[item][value])
            at[item][value] = {};
        at = at[item][value];
    }
    if (at._) {
        console.error(`Model for ${JSON.stringify(args)} already exists`);
        throw 'Already Exists';
    }
    var record = {args, getters:{}, results:{}};
    at._ = record;
    this.builds.push(record);
    log('[model] Built new model for', path);
    this.checkBuilds();
}

function get(args, filter=x=>true) {
    var path = Object.keys(args).sort();
    var at = this.buildMap;
    for (let item of path) {
        var value = args[item];
        if (!at[item] || !at[item][value])
            throw `Build for ${JSON.stringify(args)} has not been created`;
        at = at[item][value];
    }
    var build = at._;
    if (!build) {
        throw `Build for ${JSON.stringify(args)} has not been created`;
    }
    var send = {};
    for (var key in build.results) {
        if (!filter(this.sources[key])) {
            continue;
        }
        if (this.sources[key].omit) {
            continue;
        }
        send[key] = build.results[key];
    }
    return send;
}

module.exports = function(model) {
    if (model.log) log = console.log;
    model.defaultOptions = {}
    model.sources =  {}
    model.builds = []
    model.buildMap = {}
    model.add = add.bind(model);
    model.union = union.bind(model);
    model.build = build.bind(model);
    model.get = get.bind(model);
    model.checkBuilds = checkBuilds.bind(model);
    model.checkNameExists = checkNameExists.bind(model);
    model.makeGetter = makeGetter.bind(model);
    model.formUnion = formUnion.bind(model);
    return model;
};
